/*
 * shader.cpp
 *
 *  Created on: 26 mar 2016
 *      Author: crm
 */

#include "shader.h"

#include <vector>
#include <gl/glu.h>

#include "resource.h"
#include "log.h"

using namespace Engine;

bool Shader::load(const std::string& path)
	{
	const std::string pathV=path+".vert";
	const std::string pathF=path+".frag";

	clear();

	Log::debug("Shader.load: Ladowanie \"%s\" oraz \"%s\"", pathV.c_str(), pathF.c_str());

	char *data;
	GLuint err;

	GLint cresult;
	GLint cinfolength;
	const GLchar *source[1];

	/***** Vertex Shader *****/
	data=Resource::load(pathV);

	if(!data)
		{
		Log::fail("Shader.load: Nie udalo sie wczytac \"%s\"", pathV.c_str());
		return false;
		}
	else
		Log::debug("Shader.load: Wczytano kod vertex shadera");

	source[0]=data;

	GLuint shaderV=glCreateShader(GL_VERTEX_SHADER);
	glShaderSource(shaderV, 1, source, nullptr);
	glCompileShader(shaderV);

	delete [] data;

	glGetShaderiv(shaderV, GL_COMPILE_STATUS, &cresult);
	glGetShaderiv(shaderV, GL_INFO_LOG_LENGTH, &cinfolength);

	if(cresult!=GL_TRUE)
		{
		std::vector<char> info;
		info.resize(cinfolength);

		glGetShaderInfoLog(shaderV, cinfolength, NULL, info.data());

		Log::fail("Shader.load: Nie udalo sie skompilowac vertex shadera \"%s\"", pathV.c_str());
		Log::fail("Shader.load: %s", info.data());

		return 0;
		}
	else
		Log::debug("Shader.load: Vertex shader skompilowany poprawnie");

	/***** Fragment Shader *****/
	data=Resource::load(pathF);

	if(!data)
		{
		Log::fail("Shader.load: Nie udalo sie wczytac \"%s\"", pathF.c_str());
		return false;
		}
	else
		Log::debug("Shader.load: Wczytano kod fragment shadera");

	source[0]=data;

	GLuint shaderF=glCreateShader(GL_FRAGMENT_SHADER);
	glShaderSource(shaderF, 1, source, nullptr);
	glCompileShader(shaderF);

	delete [] data;

	glGetShaderiv(shaderF, GL_COMPILE_STATUS, &cresult);
	glGetShaderiv(shaderF, GL_INFO_LOG_LENGTH, &cinfolength);

	if(cresult!=GL_TRUE)
		{
		std::vector<char> info;
		info.resize(cinfolength);

		glGetShaderInfoLog(shaderF, cinfolength, NULL, info.data());

		Log::fail("Shader.load: Nie udalo sie skompilowac fragment shadera \"%s\"", pathF.c_str());
		Log::fail("Shader.load: %s", info.data());

		return 0;
		}
	else
		Log::debug("Shader.load: Fragment shader skompilowany poprawnie");

	/***** Linkowanie programu *****/
	program=glCreateProgram();
	glAttachShader(program, shaderV);
	glAttachShader(program, shaderF);
	glLinkProgram(program);

	glDetachShader(program, shaderV);
	glDetachShader(program, shaderF);

	glDeleteShader(shaderV);
	glDeleteShader(shaderF);

	if((err=glGetError())!=GL_NO_ERROR)
		{
		Log::fail("Shader.load: Linkowanie programu nie powiodlo sie \"%s\", \"%s\"", pathV.c_str(), pathF.c_str());
		Log::fail("Shader.load: %s", gluErrorString(err));

		program=0u;

		return false;
		}
	else
		Log::ok("Shader.load: Shader gotowy");

	return true;
	}

void Shader::clear()
	{
	Log::debug("Shader.clear: Czyszczenie shadera: [program: %u] [uniforms: %u] [uvalues: %u]", program, uniforms.size(), uvalues.size());

	if(program!=0u)
		glDeleteProgram(program);
	uniforms.clear();
	uvalues.clear();
	}

GLuint Shader::getUniform(const std::string& name)
	{
	auto it=uniforms.find(name);

	if(it!=uniforms.end())
		return it->second;

	const GLuint id=glGetUniformLocation(program, name.c_str());

	uniforms[name]=id;

	return id;
	}

void Shader::setUniforms()
	{
	for(auto pair: uvalues)
		{
		Uniform &u=pair.second;

		switch(u.type)
			{
			case Uniform::Type::INT:
				glUniform1i(getUniform(pair.first), u.value.uint);
			break;

			case Uniform::Type::FLOAT:
				glUniform1f(getUniform(pair.first), u.value.ufloat);
			break;

			default:
				// noop
			break;
			}
		}
	}
