/*
 * level.cpp
 *
 *  Created on: 19 kwi 2016
 *      Author: crm
 */

#include "level.h"

#include <limits>
#include <sstream>
#include <fstream>
#include <tinyxml.h>

#include "log.h"
#include "parser.h"
#include "resource.h"

#include "../consts.h"
#include "../engine/render/render.h"
#include "game.h"
#include "objecttemplate.h"
#include "backgroundobject.h"
#include "luafuncs.h"

using namespace Game;

bool Level::load(const std::string& path)
	{
	const char XML_TAG_LEVEL[]=		"level";
	const char XML_TAG_GRAVITY[]=	"gravity";
	const char XML_TAG_TEMPLATES[]=	"templates";
	const char XML_TAG_TEMPLATE[]=	"template";
	const char XML_TAG_OBJECTS[]=	"objects";
	const char XML_TAG_OBJECT[]=	"object";
	const char XML_TAG_COLLIDERS[]=	"colliders";
	const char XML_TAG_BOX[]=		"box";
	const char XML_TAG_NPCS[]=		"npcs";
	const char XML_TAG_NPC[]=		"npc";
	const char XML_TAG_PLAYER[]=	"player";
	const char XML_ATR_ID[]=		"id";
	const char XML_ATR_X[]=			"x";
	const char XML_ATR_Y[]=			"y";
	const char XML_ATR_Z[]=			"z";
	const char XML_ATR_W[]=			"w";
	const char XML_ATR_H[]=			"h";
	const char XML_ATR_FRICTION[]=	"friction";
	const char XML_ATR_SKYBOX[]=	"skybox";
	const char XML_ATR_PATH[]=		"path";
	const char XML_ATR_TEMPLATE[]=	"template";
	const char XML_ATR_BACKGROUND[]="background";
	const char XML_ATR_UNMOVING[]=  "unmoving";
	const char XML_ATR_DISABLED[]=	"disabled";
	const char XML_ATR_SCRIPT[]=	"script";
	const char XML_ATR_LUA[]=		"lua";
	const char XML_ATR_DIRECTION[]=	"direction";
	const char XML_ATR_ANIMATION[]=	"animation";

	char *data;

	clear();

	TiXmlDocument xml("");
	Log::info("Level.load: Wczytywanie poziomu \"%s\"", path.c_str());

	data=Resource::load(path);
	if(!data)
		{
		Log::fail("Level.load: Nie udalo sie wczytac pliku \"%s\"", path.c_str());
		return false;
		}

	Log::debug("Level.load: Wczytano plik \"%s\"", path.c_str());

	xml.Parse(data);
	delete [] data;

	/*************************************************************************/
	/**************************** Poziom *************************************/
	/*************************************************************************/
	TiXmlNode *xlvl=xml.FirstChild(XML_TAG_LEVEL);

	if(!xlvl)
		{
		Log::fail("Level.load: Nie znaleziono tagu \"%s\" w \"%s\"", XML_TAG_LEVEL, path.c_str());
		return false;
		}

	const char *askybox=xlvl->ToElement()->Attribute(XML_ATR_SKYBOX);

	if(!askybox)
		{
		Log::fail("Level.load: Brak skyboxa w \"%s\"", path.c_str());
		// return false;
		}
	else
		{
		Log::info("Level.load: Skybox \"%s\" w \"%s\"", askybox, path.c_str());
		skybox.setSprite(askybox);
		}

	const char *alvlid=xlvl->ToElement()->Attribute(XML_ATR_ID);

	if(!alvlid)
		{
		Log::fail("Level.load: Brak ID poziomu w \"%s\"", path.c_str());
		return false;
		}
	else
		{
		Log::info("Level.load: ID poziomu \"%s\" w \"%s\"", alvlid, path.c_str());
		id=alvlid;
		}

	const char *alvllscript=xlvl->ToElement()->Attribute(XML_ATR_SCRIPT);

	if(!alvllscript)
		{
		Log::fail("Level.load: Poziom nie posiada skryptu w \"%s\"", path.c_str());
		}
	else
		{
		Log::info("Level.load: Skrypt \"%s\" w \"%s\"", alvllscript, path.c_str());
		lpath=alvllscript;
		}

	/*************************************************************************/
	/**************************** Grawitacja *********************************/
	/*************************************************************************/
	TiXmlNode *xgrav=xlvl->ToElement()->FirstChild(XML_TAG_GRAVITY);

	if(!xgrav)
		{
		Log::fail("Level.load: Nie znaleziono tagu \"%s\" w \"%s\"", XML_TAG_GRAVITY, path.c_str());

		gravity=AVector(0, -320);
		}
	else
		{
		double x=0.0;
		double y=0.0;

		xgrav->ToElement()->Attribute(XML_ATR_X, &x);
		xgrav->ToElement()->Attribute(XML_ATR_Y, &y);

		Log::info("Level.load: Grawitacja [%.2f, %.2f] w \"%s\"", x, y, path.c_str());

		gravity=AVector(x, y);
		}

	/*************************************************************************/
	/**************************** Szablony ***********************************/
	/*************************************************************************/
	Log::info("Level.load: Wczytywanie szablonow obiektow w \"%s\"", path.c_str());

	TiXmlNode *xtpls=xlvl->FirstChild(XML_TAG_TEMPLATES);

	if(!xtpls)
		{
		Log::fail("Level.load: Nie znaleziono tagu \"%s\" w \"%s\"", XML_TAG_TEMPLATES, path.c_str());
		return false;
		}

	TiXmlNode *xtpl=xtpls->ToElement()->FirstChild(XML_TAG_TEMPLATE);

	while(xtpl)
		{
		const char *atpath=xtpl->ToElement()->Attribute(XML_ATR_PATH);

		if(!atpath)
			{
			Log::fail("Level.load: Nie znaleziono atrybutu \"%s\" tagu \"%s\" w \"%s\"", XML_ATR_PATH, XML_TAG_TEMPLATE, path.c_str());

			xtpl=xtpl->NextSibling(XML_TAG_TEMPLATE);
			continue;
			}

		if(!addTemplate(atpath))
			{
			Log::fail("Level.load: Nie udalo sie wczytac szablonu \"%s\" w \"%s\"", atpath, path.c_str());
			}
		else
			{
			Log::ok("Level.load: Dodano szablon \"%s\" w \"%s\"", atpath, path.c_str());
			}

		xtpl=xtpl->NextSibling(XML_TAG_TEMPLATE);
		}

	/*************************************************************************/
	/**************************** Obiekty ************************************/
	/*************************************************************************/
	Log::info("Level.load: Wczytywanie obiektow w \"%s\"", path.c_str());

	TiXmlNode *xobjs=xlvl->ToElement()->FirstChild(XML_TAG_OBJECTS);

	if(!xobjs)
		{
		Log::fail("Level.load: Nie znaleziono tagu \"%s\" w \"%s\"", XML_TAG_OBJECTS, path.c_str());
		return false;
		}

	TiXmlElement *xobj=xobjs->ToElement()->FirstChildElement(XML_TAG_OBJECT);

	while(xobj)
		{
		double x=0.0;
		double y=0.0;
		double z=0.0;

		bool colliders=true;

		const char *tname=nullptr;

		if(!xobj->Attribute(XML_ATR_X, &x))
			{
			Log::fail("Level.load: Nie znaleziono atrybutu \"%s\" tagu \"%s\" w \"%s\"", XML_ATR_X, XML_TAG_OBJECT, path.c_str());
			return false;
			}

		if(!xobj->Attribute(XML_ATR_Y, &y))
			{
			Log::fail("Level.load: Nie znaleziono atrybutu \"%s\" tagu \"%s\" w \"%s\"", XML_ATR_Y, XML_TAG_OBJECT, path.c_str());
			return false;
			}

		if(!xobj->Attribute(XML_ATR_Z, &z))
			{
			Log::fail("Level.load: Nie znaleziono atrybutu \"%s\" tagu \"%s\" w \"%s\"", XML_ATR_Z, XML_TAG_OBJECT, path.c_str());
			//return false;
			}

		tname=xobj->Attribute(XML_ATR_TEMPLATE);

		if(!tname)
			{
			Log::fail("Level.load: Nie znaleziono atrybutu \"%s\" tagu \"%s\" w \"%s\"", XML_ATR_TEMPLATE, XML_TAG_OBJECT, path.c_str());
			return false;
			}

		colliders=!xobj->Attribute(XML_ATR_BACKGROUND);

		Log::info("Level.load: Dodawanie obiektu z szablonu \"%s\" w \"%s\"", tname, path.c_str());

		BackgroundObject *obj=create<BackgroundObject>(tname, colliders);

		if(!obj)
			{
			delete obj;

			Log::fail("Level.load: Nie udalo sie stworzyc obiektu z szablonu \"%s\" w \"%s\"", tname, path.c_str());
			return false;
			}

		if(!addBackgroundObject(obj, AVector(x, y, z)))
			{
			delete obj;

			Log::fail("Level.load: Nie udalo sie dodac obiektu z szablonu \"%s\" w \"%s\"", tname, path.c_str());
			return false;
			}

		Log::ok("Level.load: Dodano obiekt z szablonu \"%s\" w \"%s\"", tname, path.c_str());

		xobj=xobj->NextSiblingElement(XML_TAG_OBJECT);
		}

	/*************************************************************************/
	/**************************** Collidery **********************************/
	/*************************************************************************/
	Log::info("Level.load: Wczytywanie colliderow w \"%s\"", path.c_str());

	TiXmlNode *xcols=xlvl->ToElement()->FirstChild(XML_TAG_COLLIDERS);

	if(!xcols)
		{
		Log::fail("Level.load: Nie znaleziono tagu \"%s\" w \"%s\"", XML_TAG_COLLIDERS, path.c_str());
		return false;
		}

	TiXmlElement *xbox=xcols->ToElement()->FirstChildElement(XML_TAG_BOX);

	while(xbox)
		{
		double x=0.0;
		double y=0.0;
		double w=0.0;
		double h=0.0;
		double friction=1.0;

		if(!xbox->Attribute(XML_ATR_X, &x))
			{
			Log::fail("Level.load: Nie znaleziono atrybutu \"%s\" tagu \"%s\" w \"%s\"", XML_ATR_X, XML_TAG_BOX, path.c_str());
			return false;
			}

		if(!xbox->Attribute(XML_ATR_Y, &y))
			{
			Log::fail("Level.load: Nie znaleziono atrybutu \"%s\" tagu \"%s\" w \"%s\"", XML_ATR_Y, XML_TAG_BOX, path.c_str());
			return false;
			}

		if(!xbox->Attribute(XML_ATR_W, &w))
			{
			Log::fail("Level.load: Nie znaleziono atrybutu \"%s\" tagu \"%s\" w \"%s\"", XML_ATR_W, XML_TAG_BOX, path.c_str());
			return false;
			}

		if(!xbox->Attribute(XML_ATR_H, &h))
			{
			Log::fail("Level.load: Nie znaleziono atrybutu \"%s\" tagu \"%s\" w \"%s\"", XML_ATR_H, XML_TAG_BOX, path.c_str());
			return false;
			}

		if(!xbox->Attribute(XML_ATR_FRICTION, &friction))
			{
			Log::fail("Level.load: Nie znaleziono atrybutu \"%s\" tagu \"%s\" w \"%s\"", XML_ATR_FRICTION, XML_TAG_BOX, path.c_str());
			//return false;
			}

		addCollider(Collider(AVector(x, y), AVector(w, h), friction));

		xbox=xbox->NextSiblingElement(XML_TAG_BOX);
		}

	/*************************************************************************/
	/**************************** NPC ****************************************/
	/*************************************************************************/
	Log::info("Level.load: Wczytywanie NPC w \"%s\"", path.c_str());

	TiXmlNode *xnpcs=xlvl->ToElement()->FirstChild(XML_TAG_NPCS);

	if(!xnpcs)
		{
		Log::fail("Level.load: Nie znaleziono tagu \"%s\" w \"%s\"", XML_TAG_NPCS, path.c_str());
		return false;
		}

	TiXmlElement *xnpc=xnpcs->ToElement()->FirstChildElement(XML_TAG_NPC);

	while(xnpc)
		{
		double x=0.0;
		double y=0.0;
		double z=0.0;
		double animation=0.0;

		bool colliders=true;
		bool unmoving=false;
		bool disabled=false;

		const char *tname=nullptr;
		const char *id=nullptr;
		const char *script=nullptr;
		const char *lua=nullptr;

		unsigned direction=0;

		if(!xnpc->Attribute(XML_ATR_X, &x))
			{
			Log::fail("Level.load: Nie znaleziono atrybutu \"%s\" tagu \"%s\" w \"%s\"", XML_ATR_X, XML_TAG_NPC, path.c_str());
			return false;
			}

		if(!xnpc->Attribute(XML_ATR_Y, &y))
			{
			Log::fail("Level.load: Nie znaleziono atrybutu \"%s\" tagu \"%s\" w \"%s\"", XML_ATR_Y, XML_TAG_NPC, path.c_str());
			return false;
			}

		if(!xnpc->Attribute(XML_ATR_Z, &z))
			{
			Log::fail("Level.load: Nie znaleziono atrybutu \"%s\" tagu \"%s\" w \"%s\"", XML_ATR_Z, XML_TAG_NPC, path.c_str());
			//return false;
			}

		if(!xnpc->Attribute(XML_ATR_ANIMATION, &animation))
			{
			Log::fail("Level.load: Nie znaleziono atrybutu \"%s\" tagu \"%s\" w \"%s\"", XML_ATR_ANIMATION, XML_TAG_NPC, path.c_str());
			return false;
			}

		if(!(tname=xnpc->Attribute(XML_ATR_TEMPLATE)))
			{
			Log::fail("Level.load: Nie znaleziono atrybutu \"%s\" tagu \"%s\" w \"%s\"", XML_ATR_TEMPLATE, XML_TAG_NPC, path.c_str());
			return false;
			}

		if(!(id=xnpc->Attribute(XML_ATR_ID)))
			{
			Log::fail("Level.load: Nie znaleziono atrybutu \"%s\" tagu \"%s\" w \"%s\"", XML_ATR_ID, XML_TAG_NPC, path.c_str());
			return false;
			}

		if(!(script=xnpc->Attribute(XML_ATR_SCRIPT)))
			{
			Log::fail("Level.load: Nie znaleziono atrybutu \"%s\" tagu \"%s\" w \"%s\"", XML_ATR_SCRIPT, XML_TAG_NPC, path.c_str());
			return false;
			}

		if(!(lua=xnpc->Attribute(XML_ATR_LUA)))
			{
			Log::fail("Level.load: Nie znaleziono atrybutu \"%s\" tagu \"%s\" w \"%s\"", XML_ATR_LUA, XML_TAG_NPC, path.c_str());
			return false;
			}

		if(!xnpc->Attribute(XML_ATR_DIRECTION))
			{
			Log::fail("Level.load: Nie znaleziono atrybutu \"%s\" tagu \"%s\" w \"%s\"", XML_ATR_DIRECTION, XML_TAG_NPC, path.c_str());
			return false;
			}
		else
			direction=(unsigned)(strcmp(xnpc->Attribute(XML_ATR_DIRECTION), "left")==0?Engine::Sprite::DirectionName::LEFT:Engine::Sprite::DirectionName::RIGHT);

		colliders=!xnpc->Attribute(XML_ATR_BACKGROUND);
		unmoving=!!xnpc->Attribute(XML_ATR_UNMOVING);
		disabled=!!xnpc->Attribute(XML_ATR_DISABLED);

		Log::info("Level.load: Dodawanie NPC o ID \"%s\" z szablonu \"%s\" w \"%s\"", id, tname, path.c_str());

		NPC *npc=create<NPC>(tname, colliders);

		if(!npc)
			{
			Log::fail("Level.load: Nie udalo sie stworzyc NPC z szablonu \"%s\" w \"%s\"", tname, path.c_str());
			return false;
			}

		if(game)
			{
			// ID_poziomu:ID_npc
			std::string nid=this->id+":"+id;

			Log::debug("Level.load: Szukanie NPC \"%s\" w bazie danych", nid.c_str());

			std::string ninfo=game->getDatabase().getStr(nid);

			if(ninfo.length()<1)
				{
				Log::info("Level.load: Nie znaleziono NPC \"%s\" w bazie danych, wpisywanie informacji", nid.c_str());

				// disabled x y z colliders sprite animation direction
				std::stringstream ss;
				ss << disabled << " " << x << " " << y << " " << z << " " << colliders << " " << npc->getSprite()->getName() << " " << animation << " " << direction;
				ninfo=ss.str();

				Log::debug("Level.load: Wpisywanie NPC \"%s\" do bazy danych: [disabled: %d][colliders: %d][sprite: %s][animation: %d][direction: %d]",
					nid.c_str(), disabled, colliders, npc->getSprite()->getName().c_str(), animation, direction);

				game->getDatabase().setStr(nid, ninfo);
				}
			else
				{
				Parser pars(ninfo, " \t");

				if(pars.count()<8)
					{
					Log::fail("Level.load: Brak wszystkich wymaganych informacji o NPC \"%s\" w bazie", nid.c_str());
					}
				else
					{
					disabled=pars.val(0);
					x=pars.valFloat(1);
					y=pars.valFloat(2);
					z=pars.valFloat(3);
					colliders=pars.val(4);
					npc->setSprite(pars[5]);
					animation=pars.val(6);
					direction=pars.val(7);
					}

				Log::debug("Level.load: Znaleziono NPC \"%s\" w bazie danych: [disabled: %d][colliders: %d][sprite: %s][animation: %d][direction: %d]",
					nid.c_str(), disabled, colliders, pars[5].c_str(), animation, direction);
				}
			}

		if(!addNPC(npc, AVector(x, y, z)))
			{
			delete npc;

			Log::fail("Level.load: Nie udalo sie dodac NPC z szablonu \"%s\" w \"%s\"", tname, path.c_str());
			return false;
			}

		if(!npc->init(id, script, lua))
			{
			delete npc;

			Log::fail("Level.load: Nie udalo sie zainicjalizowac NPC z szablonu \"%s\" w \"%s\"", tname, path.c_str());
			return false;
			}

		npc->setAnimation((unsigned)animation);
		npc->setAnimationDirection(direction);
		npc->setUnmoving(unmoving);
		npc->setDisabled(disabled);

		Log::ok("Level.load: Dodano NPC z szablonu \"%s\" w \"%s\"", tname, path.c_str());

		xnpc=xnpc->NextSiblingElement(XML_TAG_NPC);
		}

	/*************************************************************************/
	/**************************** Player *************************************/
	/*************************************************************************/
	Log::info("Level.load: Wczytywanie pozycji startowej gracza w \"%s\"", path.c_str());

	TiXmlNode *xplayer=xlvl->ToElement()->FirstChild(XML_TAG_PLAYER);

	if(!xplayer)
		{
		Log::fail("Level.load: Nie znaleziono tagu \"%s\" w \"%s\"", XML_TAG_PLAYER, path.c_str());
		return false;
		}
	else
		{
		double apx=0.0;
		double apy=0.0;
		double apz=0.0;
		const char *atpl=nullptr;

		xplayer->ToElement()->Attribute(XML_ATR_X, &apx);
		xplayer->ToElement()->Attribute(XML_ATR_Y, &apy);
		if(!xplayer->ToElement()->Attribute(XML_ATR_Z, &apz))
			apz=0.0;

		startpos=AVector(apx, apy, apz);

		Log::debug("Level.load: Pozycja startowa gracza: [%.2f, %.2f, %.2f] w \"%s\"", startpos.x, startpos.y, startpos.z, path.c_str());

		if(!(atpl=xplayer->ToElement()->Attribute(XML_ATR_TEMPLATE)))
			{
			Log::fail("Level.load: Nie podano szablonu gracza w \"%s\"", path.c_str());
			return false;
			}

		Log::info("Level.load: Dodawanie gracza z szablonu \"%s\" w \"%s\"", atpl, path.c_str());

		Player *player=create<Player>(atpl, false);

		if(!player)
			{
			Log::fail("Level.load: Nie udalo sie stworzyc gracza z szablonu \"%s\" w \"%s\"", atpl, path.c_str());
			return false;
			}

		if(!player->init())
			{
			Log::fail("Level.load: Nie udalo sie zainicjowac gracza w \"%s\"", path.c_str());
			return false;
			}

		if(!addPlayer(player, startpos))
			{
			Log::fail("Level.load: Nie udalo sie dodac gracza do poziomu w \"%s\"", path.c_str());
			return false;
			}

		Log::ok("Level.load: Dodano gracza z szablonu \"%s\" w \"%s\"", atpl, path.c_str());
		}

	Log::info("Level.load: Uruchamianie skryptu \"%s\" w \"%s\"", lpath.c_str(), path.c_str());

	lscript.init();
	if(lpath.empty())
		{
		// noop
		}
	else if(!lscript.open(lpath))
		{
		Log::fail("Level.load: Nie udalo sie uruchomic skryptu \"%s\" w \"%s\"", lpath.c_str(), path.c_str());
		}
	else
		{
		lscript.addRegistry(LUA_PTR_GAME, game);
		lscript.addRegistry(LUA_PTR_LEVEL, this);

		LuaFuncs::load(lscript, LuaFuncs::LEVEL);

		lscript.call(LUA_INIT);
		}

	/*************************************************************************/
	/**************************** Koniec *************************************/
	/*************************************************************************/
	this->path=path;

	xml.Clear();

	return true;
	}


bool Level::save(const std::string& path)
	{
	std::fstream out;

	out.open(path, std::ios::out|std::ios::binary);

	if(!out.is_open())
		{
		Log::fail("Level.save: Nie udalo sie otworzyc pliku \"%s\"", path.c_str());
		return false;
		}

	Log::info("Level.save: Zapisywanie do pliku \"%s\"", path.c_str());

	out << "<?xml version=\"1.0\" encoding=\"UTF-8\" ?>\n";
	if(skybox)
		out << "<level id=\"" << id.c_str() << "\" skybox=\"" << skybox->getName().c_str() << "\" script=\"" << lpath.c_str() << "\">\n";

	out << "\t<gravity x=\"" << gravity.x << "\" y=\"" << gravity.y << "\" />\n";

	// Szablony
	Log::info("Level.save: Wpisywanie szablonow obiektow do \"%s\" [%u]", path.c_str(), templates.size());

	out << "\t<templates>\n";

	for(auto p: templates)
		{
		ObjectTemplate *tobj=p.second;

		out << "\t\t<template path=\"" << tobj->getPath() << "\" />\n";
		}

	out << "\t</templates>\n";

	// Obiekty tla
	Log::info("Level.save: Wpisywanie obiektow tla do \"%s\" [%u]", path.c_str(), bgobjs.size());

	out << "\t<objects>\n";

	for(auto o: bgobjs)
		{
		out << "\t\t<object template=\"" << o->getTemplate()->getName().c_str() << "\" x=\"" << o->getPosition().x << "\" y=\"" << o->getPosition().y << "\" z=\"" << o->getPosition().z << "\" ";
		if(!o->hasColliders())
			out << "background=\"background\" ";
		out << "/>\n";
		}

	out << "\t</objects>\n";

	// Collidery
	Log::info("Level.save: Wpisywanie colliderow do \"%s\" [%u]", path.c_str(), colliders.size());

	out << "\t<colliders>\n";

	for(auto& c: colliders)
		{
		out << "\t\t<box x=\"" << c.pos.x << "\" y=\"" << c.pos.y << "\" w=\"" << c.size.x << "\" h=\"" << c.size.y << "\" friction=\"" << c.getFriction() << "\" />\n";
		}

	out << "\t</colliders>\n";

	// NPC
	Log::info("Level.save: Wpisywanie npc do \"%s\" [%u]", path.c_str(), npcs.size());

	out << "\t<npcs>\n";

	for(auto npc: npcs)
		{
		out << "\t\t<npc template=\"" << npc->getTemplate()->getName() << "\" x=\"" << npc->getPosition().x << "\" y=\"" << npc->getPosition().y << "\" z=\"" << npc->getPosition().z << "\"";
		if(!npc->hasColliders())
			out << " background=\"background\"";
		out << "\n";
		out << "\t\t     id=\"" << npc->getID().c_str() << "\" script=\"" << npc->getScriptPath().c_str() << "\" lua=\"" << npc->getLuaPath().c_str() << "\"\n";
		out << "\t\t     direction=\"" << ((npc->getAnimationDirection()==(unsigned)Engine::Sprite::DirectionName::LEFT)?"left":"right") << "\" ";
		out << "animation=\"" << npc->getAnimation() << "\" ";
		if(npc->isUnmoving())
			out << "unmoving=\"unmoving\" ";
		if(npc->isDisabled())
			out << "disabled=\"disabled\" ";
		out << "/>\n";
		}

	out << "\t</npcs>\n";

	// Gracz
	out << "\t<player template=\"" << player->getTemplate()->getName().c_str() << "\" x=\"" << startpos.x << "\" y=\"" << startpos.y << "\" z=\"" << startpos.z << "\" />\n";

	out << "</level>\n";

	if(!out.good())
		{
		Log::fail("Level.save: Nie udalo sie zapisac poziomu \"%s\"", path.c_str());
		return false;
		}

	out.close();

	Log::ok("Level.save: Zapisano poziom \"%s\"", path.c_str());

	return true;
	}


void Level::update(float dt)
	{
	if(skybox)
		skybox.update(dt);

	for(auto obj: objects)
		{
		obj->update(dt);
		}

	for(std::list<Particle*>::iterator it=particles.begin(); it!=particles.end(); ++it)
		{
		Particle *part=*it;

		part->update(dt);

		if(part->isDead())
			{
			delete part;
			particles.erase(it--);
			}
		}

	if(!lpath.empty())
		{
		lscript.addGlobal("dt", dt);
		lscript.call(LUA_UPDATE);
		}
	}

void Level::print()
	{
	if(skybox)
		Engine::Render::skybox(skybox);

	const AVector cp=game->getCamera().getPos();

	for(auto obj: objects)
		{
		const AVector op=obj->getPosition();

		if(op.x<cp.x-384 || op.x>cp.x+384 || op.y<cp.y-304 || op.y>cp.y+304)
			continue;

		obj->print();
		}

	for(auto part: particles)
		{
		part->print();
		}

	/*Engine::Render::setColor(AVector(1, 0, 1, 0.5f));
	for(auto& c: colliders)
		{
		Engine::Render::drawRectLine(c.pos+AVector(0, 0, 1), c.pos+c.size);
		}
	Engine::Render::setColor(AVector(1, 1, 1, 1));*/
	}

void Level::clear()
	{
	for(Object *obj: objects)
		delete obj;

	for(Particle *p: particles)
		delete p;

	objects.clear();
	bgobjs.clear();
	npcs.clear();
	particles.clear();

	// Player skasowany razem z obiektami
	player=nullptr;

	colliders.clear();
	skybox.clear();

	id="";
	path="";
	lpath="";

	lscript.clear();

	for(auto tobj: templates)
		delete tobj.second;
	templates.clear();
	}


char* Level::serialize(int &size)
	{
	return nullptr;
	}

void Level::deserialize(char *buff)
	{
	//
	}


Collider Level::test(const Box& b) const
	{
	for(auto& c: colliders)
		{
		if(c.collision(b))
			return c;
		}

	for(auto o: objects)
		{
		if(!o->getBoundingBox().collision(b))
			continue;

		const Collider r=o->test(b);

		if(r)
			return r;
		}

	return Collider();
	}


bool Level::addTemplate(const std::string& path)
	{
	ObjectTemplate *tobj;

	try
		{
		tobj=new ObjectTemplate();
		}
	catch(std::bad_alloc& e)
		{
		Log::fail("Level.addTemplate: Nie udalo sie stworzyc nowego obiektu: %s", e.what());
		return false;
		}

	if(!tobj->init(path))
		{
		Log::fail("Level.addTemplate: Nie udalo sie wczytac szablonu obiektu \"%s\"", path.c_str());

		delete tobj;
		return false;
		}

	if(templates.find(tobj->getName())!=templates.end())
		{
		Log::info("Level.addTemplate: Szablon obiektu \"%s\" juz wczytany \"%s\"", tobj->getName().c_str(), path.c_str());

		delete tobj;
		return true;
		}

	templates[tobj->getName()]=tobj;

	return true;
	}

bool Level::addTemplate(ObjectTemplate *tobj)
	{
	if(!tobj)
		return false;

	if(templates.find(tobj->getName())!=templates.end())
		{
		Log::info("Level.addTemplate: Szablon obiekut \"%s\" juz wczytany \"%s\"", tobj->getName().c_str(), path.c_str());
		return true;
		}

	templates[tobj->getName()]=tobj;

	return true;
	}

bool Level::addObject(Object *obj, const AVector& pos)
	{
	obj->setLevel(*this);
	obj->setGame(game);
	obj->setPosition(pos);

	objects.push_back(obj);

	return true;
	}

bool Level::addBackgroundObject(BackgroundObject *obj, const AVector& pos)
	{
	if(!obj)
		return false;

	addObject(obj, pos);

	bgobjs.push_back(obj);

	return true;
	}

bool Level::addNPC(NPC *obj, const AVector& pos)
	{
	if(!obj)
		return false;
	if(getNPCByID(obj->getID()))
		{
		Log::info("Level.addNPC: NPC po ID \"%s\" juz jest dodany", obj->getID().c_str());
		return false;
		}

	addObject(obj, pos);

	npcs.push_back(obj);

	return obj;
	}

bool Level::addPlayer(Player *obj, const AVector& pos)
	{
	if(!obj)
		return false;

	addObject(obj, pos);

	player=obj;

	return true;
	}

bool Level::addParticle(Particle *obj, const AVector& pos)
	{
	if(!obj)
		return false;

	particles.push_front(obj);

	obj->setPosition(pos);
	obj->setLevel(this);

	return true;
	}


void Level::addCollider(const Collider& c)
	{
	colliders.push_back(c);
	}

void Level::setGame(Game *s)
	{
	game=s;

	for(auto o: objects)
		o->setGame(s);
	}


NPC* Level::getNPCByID(const std::string& id) const
	{
	for(auto npc: npcs)
		{
		if(npc->getID()==id)
			return npc;
		}

	return nullptr;
	}

NPC* Level::getNPCClosestTo(const AVector& pos, float maxDistance) const
	{
	float bestd=maxDistance;
	NPC *bestn=nullptr;

	if(bestd<0.0f)
		bestd=64.0f;

	for(auto npc: npcs)
		{
		if(npc->isDisabled())
			continue;
		if(!npc->hasDialog())
			continue;

		float d=AVectorLength(npc->getPosition()-pos).w;

		if(d<bestd)
			{
			bestd=d;
			bestn=npc;
			}
		}

	return bestn;
	}
