/*
 * lua2.cpp
 *
 *  Created on: 02-10-2013
 *      Author: crm
 */

#include "lua2.h"

#ifndef LINUX
	#include <lua/lua.hpp>
	#include <lua/lualib.h>
	#include <lua/lauxlib.h>
#else
	#include <lua.hpp>
	#include <lualib.h>
	#include <lauxlib.h>
#endif

#include <sstream>	// tylko do createFunction

#include "log.h"
#include "parser.h"
#include "resource.h"

#include <fstream>

//void stackdump(lua_State *L, int line);
void stackdump(lua_State *L, int line) {
      fprintf(stdout, "% 4d [%d]: ", line, lua_gettop(L));

      int i;
      int top = lua_gettop(L);
      for (i = 1; i <= top; i++) {   //repeat for each level
        int t = lua_type(L, i);
        switch (t) {

          case LUA_TSTRING:  // strings
            fprintf(stdout, "'%s'", lua_tostring(L, i));
            break;

          case LUA_TBOOLEAN:   //booleans
            fprintf(stdout, lua_toboolean(L, i) ? "true" : "false");
            break;

          case LUA_TNUMBER:  // numbers
            fprintf(stdout, "%g", lua_tonumber(L, i));
            break;

          default:   //other values
            fprintf(stdout, "%s", lua_typename(L, t));
            break;

        }
        fprintf(stdout, "  ");   //put a separator
      }
      fprintf(stdout, "\n");   //end the listing
    }


bool Lua::init()
	{
	clear();

	state=luaL_newstate();
	if(!state)
		{
		Log::fail("Lua.init: Nie udalo sie stworzyc state");

		return false;
		}

	luaL_openlibs(state);

	status=0;
	args=0;
	path="script";

	return true;
	}

void Lua::clear()
	{
	if(state)
		lua_close(state);
	state=NULL;
	//name="cleared";
	}


bool Lua::open(const std::string& name)
	{
	if(!Resource::find(name))
		{
		Log::fail("Lua.open: Nie znaleziono pliku \"%s\"", name.c_str());

		return false;
		}

	char *data;
	data=Resource::load(name);

	if(!data)
		{
		Log::fail("Lua.open: Nie udalo sie wczytac pliku \"%s\"", name.c_str());

		return false;
		}

	path=name;
	addRegistry("__PATH__", name);

	if(!parse(data))
		{
		Log::fail("Lua.open: Problem przy parsowaniu skryptu \"%s\"", name.c_str());

		delete [] data;

		return false;
		}

	delete [] data;

	return true;
	}

bool Lua::call(const std::string& name)
	{
	if(status!=0)
		return false;

	lua_getglobal(state, name.c_str());
	lua_insert(state, lua_gettop(state)-args);

	status=lua_pcallk(state, args, LUA_MULTRET, 0, 0, NULL);
	args=0;

	return error();
	}

bool Lua::call(int ref)
	{
	lua_rawgeti(state, LUA_REGISTRYINDEX, ref);

	if(!lua_isfunction(state, -1))
		{
		lua_settop(state, 0);

		return false;
		}

	lua_insert(state, lua_gettop(state)-args);	// Przesun wskaznik na funkcje na spod stosu

	status=lua_pcall(state, args, LUA_MULTRET, 0);
	args=0;

	error();
	return true;
	}

bool Lua::parse(const std::string& script)
	{
	if(status!=0)
		return false;

	/*lua_getglobal(state, "debug");        // push debug-table on the stack
	lua_getfield(state, -1, "traceback");     // push debug.traceback*/
	status=luaL_loadbufferx(state, script.c_str(), script.size(), path.c_str(), NULL);
	if(!error())
		return false;

	status=lua_pcallk(state, 0, LUA_MULTRET, 0, 0, NULL);

	return error();
	}


void Lua::push(int val)
	{
	++args;
	lua_pushinteger(state, val);
	}

void Lua::push(float val)
	{
	++args;
	lua_pushnumber(state, val);
	}

void Lua::push(const std::string& str)
	{
	++args;
	lua_pushstring(state, str.c_str());
	}

void Lua::pop()
	{
	lua_settop(state, -2);
	}

void Lua::addGlobal(const std::string& name, int val)
	{
	lua_pushstring(state, name.c_str());
	lua_pushinteger(state, val);
	lua_settable(state, LUA_REGISTRYINDEX);

	/*if(!addTableStack(name))
		return;
	lua_pushinteger(state, val);
	lua_settable(state, -3);
	lua_pop(state, 1);*/
	}

void Lua::addGlobal(const std::string& name, float val)
	{
	/*lua_pushstring(state, name.c_str());
	lua_pushnumber(state, val);
	lua_settable(state, LUA_REGISTRYINDEX);*/
	if(!addTableStack(name))
		return;
	lua_pushnumber(state, val);
	lua_settable(state, -3);
	lua_pop(state, 1);
	}

void Lua::addGlobal(const std::string& name, const std::string& str)
	{
	/*lua_pushstring(state, name.c_str());
	lua_pushstring(state, str.c_str());
	lua_settable(state, LUA_REGISTRYINDEX);*/
	if(!addTableStack(name))
		return;
	lua_pushstring(state, str.c_str());
	lua_settable(state, -3);
	lua_pop(state, 1);
	}

void Lua::addGlobal(const std::string& name, void *ptr)
	{
	/*lua_pushstring(state, name.c_str());
	lua_pushlightuserdata(state, ptr);
	lua_settable(state, LUA_REGISTRYINDEX);*/
	if(!addTableStack(name))
		return;
	lua_pushlightuserdata(state, ptr);
	lua_settable(state, -3);
	lua_pop(state, 1);
	}

void Lua::addGlobal(const std::string& name, LuaFunction fptr)
	{
	/*lua_pushcclosure(state, fptr, 0);	// n -> upvalues, 'zewnetrzne' argumenty; n -> liczba rzeczy na stosie pod fptr do popniecia
	lua_setglobal(state, name.c_str());*/

	if(!addTableStack(name))
		return;
	lua_pushcclosure(state, fptr, 0);
	lua_settable(state, -3);
	lua_pop(state, 1);
	}


void Lua::addRegistry(const std::string& name, int val)
	{
	lua_pushstring(state, name.c_str());
	lua_pushinteger(state, val);
	lua_settable(state, LUA_REGISTRYINDEX);
	}

void Lua::addRegistry(const std::string& name, float val)
	{
	lua_pushstring(state, name.c_str());
	lua_pushnumber(state, val);
	lua_settable(state, LUA_REGISTRYINDEX);
	}

void Lua::addRegistry(const std::string& name, const std::string& str)
	{
	lua_pushstring(state, name.c_str());
	lua_pushstring(state, str.c_str());
	lua_settable(state, LUA_REGISTRYINDEX);
	}

void Lua::addRegistry(const std::string& name, void *ptr)
	{
	lua_pushstring(state, name.c_str());
	lua_pushlightuserdata(state, ptr);
	lua_settable(state, LUA_REGISTRYINDEX);
	}


bool Lua::error()
	{
	switch(status)
		{
		case LUA_ERRRUN:
			Log::fail("Lua.error[%s]: Problem z wykonaniem: %s", path.c_str(), lua_tostring(state, -1));
		break;

		case LUA_ERRSYNTAX:
			Log::fail("Lua.error[%s]: Blad skladni: %s", path.c_str(), lua_tostring(state, -1));
		break;

		case LUA_ERRMEM:
			Log::fail("Lua.error[%s]: Problem z alokacja pamieci: %s", path.c_str(), lua_tostring(state, -1));
		break;

		case LUA_ERRGCMM:
			Log::fail("Lua.error[%s]: Problem z garbage collectorem: %s", path.c_str(), lua_tostring(state, -1));
		break;

		case LUA_ERRERR:
			Log::fail("Lua.error[%s]: Problem obsluga bledu: %s", path.c_str(), lua_tostring(state, -1));
		break;

		case LUA_ERRFILE:
			Log::fail("Lua.error[%s]: Problem z wczytaniem pliku: %s", path.c_str(), lua_tostring(state, -1));
		break;

		default:
		case LUA_OK:
			return true;
		break;
		}

	return false;
	}


bool Lua::addTableStack(const std::string& name)
	{
	Parser pars;

	pars.setDelimiter(".");
	pars.setText(name);

	lua_pushglobaltable(state);

	for(int i=0; i<pars.count()-1; ++i)
		{
		const std::string& t=pars[i];

		lua_pushstring(state, t.c_str());
		lua_gettable(state, -2);

		if(lua_isnil(state, -1))
			{
			lua_pop(state, 1);

			lua_createtable(state, 0, 0);
			lua_pushstring(state, t.c_str());
			lua_pushvalue(state, -2);
			lua_settable(state, -4);
			lua_remove(state, -2);
			}
		else if(lua_istable(state, -1))
			lua_remove(state, -2);
		else
			{
			lua_remove(state, -3);
			return false;
			}

		//stackdump(state, 288);
		}

	//lua_pop(state, 1);
	lua_pushstring(state, pars[pars.count()-1].c_str());

	//stackdump(state, 293);

	return true;
	}

bool Lua::functionExists(const std::string& name)
	{
	bool ret;

	lua_getglobal(state, name.c_str());

	ret=lua_isfunction(state, -1);

	lua_settop(state, 0);

	return ret;
	}

void Lua::createFunction(const std::string& name)
	{
	std::stringstream ss;
	ss << "function " << name << "() end";

	parse(ss.str());
	}

#ifdef UTILS_TEST
//#include "log.h"
void lua2_h_main()
	{
	Log::info("start");

	Lua lua;
	lua.init();

	lua.addGlobal("raz.dwa.cztery", 15);
	lua.addGlobal("raz.dwa.piec.siedem", 51);
	lua.addGlobal("raz.dwa.a.b", 105);
	lua.addGlobal("b", 501);
	//stackdump((lua_State*)(&lua), 326);
	lua.parse("print('tablica: '..raz.dwa.cztery..';');");
	lua.parse("print('tablica: '..raz.dwa.piec.siedem..';');");
	lua.parse("print('tablica: '..raz.dwa.a.b..';');");
	lua.parse("print('tablica: '..b..';');");
	fflush(stdout);

	lua.clear();
	}
#endif // UTILS_TEST

